/*
 * Copyright (C) 2019  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/domain/search_domain.hh"

#include "src/util/sql_exec_in_thread.hh"
#include "src/util/strong_type.hh"
#include "src/util/struct_to_string.hh"

#include "util/log/logger.hh"

#include <boost/uuid/string_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

#include <array>
#include <map>

namespace Fred {
namespace Registry {
namespace Domain {

std::string SearchDomainRequest::to_string() const
{
    return Util::StructToString().add("query_values", query_values)
                                 .add("limit", limit)
                                 .add("searched_items", searched_items)
                                 .finish();
}

std::string SearchDomainReply::Data::to_string() const
{
    return Util::StructToString().add("most_similar_domains", most_similar_domains)
                                 .add("result_count", result_count)
                                 .add("searched_items", searched_items)
                                 .finish();
}

std::string SearchDomainReply::Data::Result::to_string() const
{
    return Util::StructToString().add("domain_id", domain_id)
                                 .add("matches", matches)
                                 .add("matched_items", matched_items)
                                 .finish();
}

std::string SearchDomainReply::Data::Result::Match::to_string() const
{
    return Util::StructToString().add("similarity", similarity)
                                 .add("item", item)
                                 .finish();
}

std::string SearchDomainReply::Data::FuzzyValue::to_string() const
{
    return Util::StructToString().add("lower_estimate", lower_estimate)
                                 .add("upper_estimate", upper_estimate)
                                 .finish();
}

SearchDomainReply::Data search_domain(const SearchDomainRequest& request)
{
    const bool unlimited = request.limit == boost::none;
    constexpr auto number_of_searched_items = 1;
    typedef SearchDomainReply::Data::Result (*GetResultFromMatch)(const Database::Result::Row&);
    class ToResult
    {
    public:
        static SearchDomainReply::Data::Result from_handle_searching(const Database::Result::Row& row)
        {
            return from_searching(row, Domain::DomainItem::fqdn);
        }
    private:
        static SearchDomainReply::Data::Result from_searching(
                const Database::Result::Row& row,
                Domain::DomainItem item)
        {
            SearchDomainReply::Data::Result result;
            result.domain_id =
                    Util::make_strong<Registry::Domain::DomainId>(
                            boost::uuids::string_generator()(static_cast<std::string>(row[0])));
            result.matched_items.insert(item);
            SearchDomainReply::Data::Result::Match match;
            match.similarity = static_cast<float>(row[1]);
            match.item = item;
            result.matches.push_back(match);
            return result;
        }
    };
    struct Task
    {
        Task(const std::string& _sql, GetResultFromMatch _get_result_from_match, DomainItem _item)
            : sql(_sql),
              get_result_from_match(_get_result_from_match),
              item(_item)
        { }
        const std::string sql;
        const GetResultFromMatch get_result_from_match;
        const DomainItem item;
    };
    static const std::array<Task, number_of_searched_items> task =
            {
                Task(
                    "WITH unaccented AS (SELECT f_unaccent($1::TEXT) AS query) "
                    "SELECT uuid,similarity(name,unaccented.query) "
                    "FROM unaccented,object_registry "
                    "WHERE name % unaccented.query AND "
                          "type=get_object_type_id('domain') AND "
                          "erdate IS NULL "
                    "ORDER BY 2 DESC,id",
                    ToResult::from_handle_searching,
                    DomainItem::fqdn
                )
            };
    const Database::QueryParams params =
            unlimited ? Database::query_param_list(request.query_values[0])
                      : Database::query_param_list(request.query_values[0])(*request.limit);
    std::array<Util::SqlInThreadSolver, number_of_searched_items> running_task;
    for (auto idx = 0; idx < number_of_searched_items; ++idx)
    {
        const bool has_to_be_searched = 0 < request.searched_items.count(task[idx].item);
        if (has_to_be_searched)
        {
            running_task[idx] = Util::sql_exec_in_thread(
                    unlimited ? task[idx].sql
                              : task[idx].sql + " LIMIT $2::BIGINT",
                    params);
        }
    }
    std::array<Util::SqlInThreadSolver::Result, number_of_searched_items> match;
    SearchDomainReply::Data result;
    result.result_count.lower_estimate = 0;
    result.result_count.upper_estimate = 0;
    struct Column
    {
        std::size_t top;
        std::size_t end;
        DomainId domain_id;
        DomainItem item;
    };
    std::array<Column, number_of_searched_items> column;
    std::map<float, std::set<std::size_t>, std::greater<float>> most_similar;
    for (auto col_idx = 0; col_idx < number_of_searched_items; ++col_idx)
    {
        const bool was_searched = 0 < request.searched_items.count(task[col_idx].item);
        if (was_searched)
        {
            match[col_idx] = std::move(running_task[col_idx]);
            if (0 < match[col_idx].get_result().size())
            {
                result.searched_items.insert(task[col_idx].item);
                const auto matches = match[col_idx].get_result().size();
                if (result.result_count.lower_estimate < matches)
                {
                    result.result_count.lower_estimate = matches;
                }
                result.result_count.upper_estimate += matches;
                const auto item_match_result = task[col_idx].get_result_from_match(match[col_idx].get_result()[0]);
                column[col_idx].top = 0;
                column[col_idx].end = match[col_idx].get_result().size();
                column[col_idx].domain_id = item_match_result.domain_id;
                column[col_idx].item = item_match_result.matches[0].item;
                most_similar[item_match_result.matches[0].similarity].insert(col_idx);
            }
        }
    }
    const std::size_t limit = unlimited ? result.result_count.upper_estimate
                                        : *request.limit;
    if (unlimited || (result.result_count.upper_estimate <= *request.limit))
    {
        result.most_similar_domains.reserve((result.result_count.lower_estimate + result.result_count.upper_estimate) / 2);
    }
    else
    {
        result.most_similar_domains.reserve(*request.limit);
    }
    result.searched_items = request.searched_items;
    std::map<boost::uuids::uuid, std::size_t> domain_id_to_result_idx;
    while (!most_similar.empty())
    {
        const auto column_idx = *(most_similar.begin()->second.begin());
        SearchDomainReply::Data::Result::Match col_match;
        col_match.similarity = most_similar.begin()->first;
        col_match.item = column[column_idx].item;
        const auto domain_id_raw = Util::get_raw_value_from(column[column_idx].domain_id);
        const auto domain_id_to_result_idx_itr = domain_id_to_result_idx.find(domain_id_raw);
        const bool domain_in_results = domain_id_to_result_idx_itr != domain_id_to_result_idx.end();
        if (domain_in_results)
        {
            result.most_similar_domains[domain_id_to_result_idx_itr->second].matches.push_back(col_match);
            --result.result_count.upper_estimate;
        }
        else
        {
            if (result.most_similar_domains.size() < limit)
            {
                SearchDomainReply::Data::Result domain;
                domain.domain_id = column[column_idx].domain_id;
                domain.matches.push_back(col_match);
                domain.matched_items.insert(column[column_idx].item);
                domain_id_to_result_idx[domain_id_raw] = result.most_similar_domains.size();
                result.most_similar_domains.push_back(domain);
            }
        }
        most_similar.begin()->second.erase(column_idx);
        if (most_similar.begin()->second.empty())
        {
            most_similar.erase(most_similar.begin());
        }
        column[column_idx].top += 1;
        if (column[column_idx].top < column[column_idx].end)
        {
            const auto item_match_result = task[column_idx].get_result_from_match(
                    match[column_idx].get_result()[column[column_idx].top]);
            column[column_idx].domain_id = item_match_result.domain_id;
            column[column_idx].item = item_match_result.matches[0].item;
            most_similar[item_match_result.matches[0].similarity].insert(column_idx);
        }
    }
    if (result.result_count.lower_estimate < result.most_similar_domains.size())
    {
        result.result_count.lower_estimate = result.most_similar_domains.size();
    }
    if (unlimited || (result.most_similar_domains.size() < *request.limit))
    {
        if ((result.result_count.lower_estimate != result.result_count.upper_estimate) ||
            (result.result_count.lower_estimate != result.most_similar_domains.size()))
        {
            LOGGER.info("estimated number of results does not correspond with reality: "
                        "<" + std::to_string(result.result_count.lower_estimate) + ", " +
                              std::to_string(result.result_count.upper_estimate) + "> ? " +
                        std::to_string(result.most_similar_domains.size()));
            result.result_count.lower_estimate = result.most_similar_domains.size();
            result.result_count.upper_estimate = result.most_similar_domains.size();
        }
    }
    else if ((result.most_similar_domains.size() != *request.limit) ||
            (result.result_count.upper_estimate < result.result_count.lower_estimate) ||
            (result.result_count.upper_estimate < *request.limit))
    {
        LOGGER.info("estimated number of results does not correspond with reality: "
                    "<" + std::to_string(result.result_count.lower_estimate) + ", " +
                          std::to_string(result.result_count.upper_estimate) + "> ? "
                    "{" + std::to_string(result.most_similar_domains.size()) + ", " +
                          std::to_string(*request.limit) + "}");
    }
    return result;
}

} // namespace Fred::Registry::Domain
} // namespace Fred::Registry
} // namespace Fred
